/*
Critical Mass
Win32 GUI component
Copyright(C) 2002, Jem Berkes <umberkes at cc dot umanitoba dot ca>
*/


#include "defs.h"
#include "board.h"
#include "strings-en.h"
#include "resource.h"
#include <stdio.h>
#include <time.h>
#include <windows.h>
#include <vector>

using namespace std;


// Global variables
HINSTANCE myinstance;
HWND hWnd;
HBITMAP piece1, piece2, piece3;
HCURSOR pointer, hourglass;
int turn;
int p1_wins =		0;
int p2_wins =		0;
int board_width =	DEF_BOARDWIDTH;
int board_height =	DEF_BOARDHEIGHT;
int whats_player1 =	HUMAN_PLAYER;
int whats_player2 =	COMP_PLAYER;
bool intelligent =	true;
bool sound_on =		true;
int replay =		0;			// keeps track during replay mode
int last_x, last_y;				// used during replay mode
the_board* gameboard = NULL;	// the game board current displayed
vector<the_board> history;		// history of moves, used for undo


// Function prototypes
LRESULT CALLBACK handler(HWND hWnd, UINT, WPARAM, LPARAM);
LRESULT CALLBACK game_options(HWND, UINT, WPARAM, LPARAM);
void undo_one();
void paint_window(HDC);
void got_timeslice();
void valid_click(int, int);
void human_move(uchar, int, int);
void computer_move(uchar);
void winner_check();
void cleanup_all();


/*
 The main function defines start point. Initializes some values,
 loads objects, registers a window class and then creates window.
*/
int WINAPI WinMain
	(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR cmdParam, int cmdShow)
{
// Initialize values
srand((unsigned)time(NULL));
myinstance = hInst;
piece1 = LoadBitmap(myinstance, (LPCTSTR)IDB_PIECE1);
piece2 = LoadBitmap(myinstance, (LPCTSTR)IDB_PIECE2);
piece3 = LoadBitmap(myinstance, (LPCTSTR)IDB_PIECE3);
pointer = LoadCursor(NULL, IDC_ARROW);
hourglass = LoadCursor(NULL, IDC_WAIT);

// Register window class
WNDCLASSEX wcex;
wcex.cbSize			= sizeof(WNDCLASSEX); 
wcex.style			= CS_HREDRAW | CS_VREDRAW;
wcex.lpfnWndProc	= (WNDPROC)handler;
wcex.cbClsExtra		= 0;
wcex.cbWndExtra		= 0;
wcex.hInstance		= hInst;
wcex.hIcon			= LoadIcon(myinstance, (LPCTSTR)IDI_ICON1);
wcex.hCursor		= pointer;
wcex.hbrBackground	= (HBRUSH)GetStockObject(WHITE_BRUSH);
wcex.lpszMenuName	= (LPCTSTR)IDR_MENU1;
wcex.lpszClassName	= WINDOW_CLASS;
wcex.hIconSm		= NULL;
RegisterClassEx(&wcex);

// Create and display the window
hWnd = CreateWindow(WINDOW_CLASS, TITLE_PROG, WS_OVERLAPPEDWINDOW,
	CW_USEDEFAULT, 0, WINDOWSIZE, WINDOWSIZE, GetDesktopWindow(),
	NULL, hInst, NULL);
ShowWindow(hWnd, SW_SHOWNORMAL);
UpdateWindow(hWnd);

// Activate timer
SetTimer(hWnd, MYTIMER_ID, DELAY_MEDIUM, NULL);

// Enter message loop
MSG  msg;
int status;
while ((status = GetMessage (&msg, 0, 0, 0)) != 0)
	{
	if (status == -1)
		return -1;
	TranslateMessage(&msg);
	DispatchMessage (&msg);
	}

cleanup_all();
return 0;
}


/*
 The main message handler for this program. Responds to various
 user and system events
*/
LRESULT CALLBACK handler(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
switch (msg)
	{
	case WM_LBUTTONDOWN:
		// USER CLICKED LEFT MOUSE BUTTON
		{
		if (gameboard!=NULL)
			{
			if ( (turn<=1) || (gameboard->who_won()==0) )
				{
				RECT rt;
				GetClientRect(hWnd, &rt);
				int x_abs = LOWORD(lParam);
				int y_abs = HIWORD(lParam);
				int x_steps = rt.right / board_width;
				int y_steps = rt.bottom / board_height;
				int xpos = x_abs / x_steps;
				int ypos = y_abs / y_steps;
				if (xpos<board_width && ypos<board_height)
					valid_click(xpos, ypos);
				}
			}
		break;
		}

	case WM_PAINT:
		// SYSTEM NEEDS TO PAINT
		if (gameboard != NULL)
			paint_window((HDC)wParam);
		break;

	case WM_DESTROY:
		// SYSTEM DEMANDS EXIT
		PostQuitMessage(0);
		return 0;

	case WM_TIMER:
		// TIMER EVENT (WE SET THE TIMER EARLIER)
		if ( (gameboard != NULL) && (wParam == MYTIMER_ID) )
			{
			if (replay != 0)		// if replay is active
				got_timeslice();	// always respond to timer
			else if ( (turn<=1) || (gameboard->who_won()==0) )
				got_timeslice();
			}
		return 0;

	case WM_CLOSE:
		// USER OR SYSTEM TRIED TO CLOSE WINDOW
		PostMessage(hWnd, WM_COMMAND, MAKEWPARAM(ID_FILE_EXIT, 0), 0);
		return 0;

	case WM_COMMAND:
		{
		switch (LOWORD(wParam))
			{
			case ID_GAME_REPLAYMOVE:
				// USER WANTS TO REPLAY LAST MOVE
				{
				if ( !history.empty() && (replay==0) &&
					gameboard->read_lastmove(last_x, last_y) )
					{
					delete gameboard;
					gameboard = new the_board(history.back());
					replay = 1;	// 1 = last board state restored
								// pause one timeslice, then continue
					paint_window(NULL);
					}
				break;
				}

			case ID_GAME_UNDO:
				// USER WANTS TO UNDO LAST MOVE MADE
				{
				if (history.empty() || (replay!=0))
					MessageBox(hWnd, ERR_CANTUNDO, NULL, MB_ICONERROR|MB_OK);
				else
					{
					undo_one();
					if ( ((turn%2==0) && (whats_player1==COMP_PLAYER)) ||
						 ((turn%2==1) && (whats_player2==COMP_PLAYER)) )
						// If it's the computer's turn, undo that as well
						undo_one();
					paint_window(NULL);
					}
				break;
				}

			case ID_GAME_STATISTICS:
				// USER WANTS TO SEE GAME STATISTICS
				{
				char msgbuf[512];
				sprintf(msgbuf, MSG_P1WINS"\t%d\n"MSG_P2WINS"\t%d\n",
					p1_wins, p2_wins);
				MessageBox(hWnd,msgbuf,TITLE_STATS,MB_ICONINFORMATION|MB_OK);
				break;
				}

			case ID_FILE_NEWGAME:
				// USER WANTS TO START NEW GAME
				if (DialogBox(myinstance, (LPCTSTR)IDD_NEWGAME, hWnd,
					(DLGPROC)game_options) == 1)
					{
					if (gameboard != NULL)
						delete gameboard;
					gameboard = new the_board(board_width, board_height);
					turn = 0;
					replay = 0;
					history.clear();
					paint_window(NULL);
					}
				break;
				
			case ID_FILE_EXIT:
				// USER WANTS TO EXIT
				if (MessageBox(hWnd, QUESTION_LEAVE, TITLE_PROG,
					MB_ICONQUESTION|MB_YESNO) == IDYES)
					DestroyWindow(hWnd);
				break;

			case ID_GAME_SLOW:
				// USER SELECTS SLOW COMPUTER SPEED
				{
				HMENU themenu = GetMenu(hWnd);
				CheckMenuItem(themenu, ID_GAME_SLOW, MF_CHECKED);
				CheckMenuItem(themenu, ID_GAME_MEDIUM, MF_UNCHECKED);
				CheckMenuItem(themenu, ID_GAME_FAST, MF_UNCHECKED);
				KillTimer(hWnd, MYTIMER_ID);
				SetTimer(hWnd, MYTIMER_ID, DELAY_SLOW, NULL);
				break;
				}

			case ID_GAME_MEDIUM:
				// USER SELECTS MEDIUM COMPUTER SPEED
				{
				HMENU themenu = GetMenu(hWnd);
				CheckMenuItem(themenu, ID_GAME_SLOW, MF_UNCHECKED);
				CheckMenuItem(themenu, ID_GAME_MEDIUM, MF_CHECKED);
				CheckMenuItem(themenu, ID_GAME_FAST, MF_UNCHECKED);
				KillTimer(hWnd, MYTIMER_ID);
				SetTimer(hWnd, MYTIMER_ID, DELAY_MEDIUM, NULL);
				break;
				}
			
			case ID_GAME_FAST:
				// USER SELECTS FAST COMPUTER SPEED
				{
				HMENU themenu = GetMenu(hWnd);
				CheckMenuItem(themenu, ID_GAME_SLOW, MF_UNCHECKED);
				CheckMenuItem(themenu, ID_GAME_MEDIUM, MF_UNCHECKED);
				CheckMenuItem(themenu, ID_GAME_FAST, MF_CHECKED);
				KillTimer(hWnd, MYTIMER_ID);
				SetTimer(hWnd, MYTIMER_ID, DELAY_FAST, NULL);
				break;
				}

			case ID_GAME_SOUND_ON:
				// USER WANTS SOUND ON
				{
				HMENU themenu = GetMenu(hWnd);
				CheckMenuItem(themenu, ID_GAME_SOUND_ON, MF_CHECKED);
				CheckMenuItem(themenu, ID_GAME_SOUND_OFF, MF_UNCHECKED);
				sound_on = true;
				break;
				}

			case ID_GAME_SOUND_OFF:
				// USER WANTS SOUND OFF
				{
				HMENU themenu = GetMenu(hWnd);
				CheckMenuItem(themenu, ID_GAME_SOUND_ON, MF_UNCHECKED);
				CheckMenuItem(themenu, ID_GAME_SOUND_OFF, MF_CHECKED);
				sound_on = false;
				break;
				}

			case ID_HELP_ABOUT:
				MessageBox(hWnd, MSG_ABOUT, TITLE_PROG, MB_OK);
				break;

			case ID_HELP_HOWTOPLAY:
				MessageBox(hWnd, MSG_HELPTEXT,
					TITLE_HELP, MB_ICONINFORMATION|MB_OK);
				break;

			case ID_HELP_INTERNETHELP:
				if ((int)ShellExecute(hWnd, "open", URL_GAME,
					NULL, NULL, SW_SHOWNORMAL) < 33)
					MessageBox(hWnd,ERR_BROWSE,TITLE_HELP,MB_ICONERROR|MB_OK);
				break;

			} // switch
		break;
		} // WM_COMMAND
	} // switch
return DefWindowProc(hWnd, msg, wParam, lParam);
}


/*
 Message handler for dialog box in which user sets game options.
 Standard windows dialog box.
*/
LRESULT CALLBACK
	game_options(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
switch (message)
	{
	case WM_INITDIALOG:
		// DIALOG BOX STARTS UP...
		// What type of beast is player1
		if (whats_player1 == HUMAN_PLAYER)
			SendDlgItemMessage(hDlg, IDC_P1_HUMAN, BM_SETCHECK,BST_CHECKED,0);
		else
			SendDlgItemMessage(hDlg, IDC_P1_COMP, BM_SETCHECK,BST_CHECKED,0);
		// What type of beast is player2
		if (whats_player2 == HUMAN_PLAYER)
			SendDlgItemMessage(hDlg, IDC_P2_HUMAN, BM_SETCHECK,BST_CHECKED,0);
		else
			SendDlgItemMessage(hDlg, IDC_P2_COMP, BM_SETCHECK,BST_CHECKED,0);
		// Set board dimensions
		SetDlgItemInt(hDlg, IDC_SIZEX, board_width, TRUE);
		SetDlgItemInt(hDlg, IDC_SIZEY, board_height, TRUE);
		// Is intelligence on
		if (intelligent)
			SendDlgItemMessage(hDlg, IDC_SMART, BM_SETCHECK, BST_CHECKED, 0);
		return TRUE;

		
	case WM_COMMAND:
		switch (LOWORD(wParam))
			{
			case IDCANCEL:
				// USER CANELS
				EndDialog(hDlg, 0);
				return TRUE;

			case IDOK:
				// USER WANTS TO "OK" THE DIALOG BOX
				// Who is player 1
				if (SendDlgItemMessage(hDlg, IDC_P1_HUMAN, BM_GETCHECK, 0, 0)
					== BST_CHECKED)
					whats_player1 = HUMAN_PLAYER;
				else
					whats_player1 = COMP_PLAYER;
				// Who is player 2
				if (SendDlgItemMessage(hDlg, IDC_P2_HUMAN, BM_GETCHECK, 0, 0)
					== BST_CHECKED)
					whats_player2 = HUMAN_PLAYER;
				else
					whats_player2 = COMP_PLAYER;
				// Check intelligence setting
				intelligent = (SendDlgItemMessage(hDlg, IDC_SMART,
					BM_GETCHECK, 0, 0) == BST_CHECKED);
				// Read board dimensions
				board_width = GetDlgItemInt(hDlg, IDC_SIZEX, NULL, TRUE);
				board_height = GetDlgItemInt(hDlg, IDC_SIZEY, NULL, TRUE);
				if ( (board_width<2) || (board_height<2) )
					MessageBox(hDlg, ERR_BOARDSIZE, NULL, MB_ICONWARNING);
				else
					EndDialog(hDlg, 1);	// 1 means user confirmed
				return TRUE;
			} // switch
		break;
	} // switch (message)
return FALSE;
}


/*
 Undo a single move
*/
void undo_one()
{
if (!history.empty())
	{
	delete gameboard;
	gameboard = new the_board(history.back());
	history.pop_back();
	turn--;
	}
}


/*
 We must draw the board on the screen. This may be called at any time.
*/
void paint_window(HDC hdc_actual)
{
RECT space;
GetClientRect(hWnd, &space);
int orig_width = space.right;
int orig_height = space.bottom;
if (hdc_actual == NULL)
	hdc_actual = GetDC(hWnd);
// Create an HDC in memory for fast drawing (no flicker)
HDC hdc = CreateCompatibleDC(hdc_actual);
HBITMAP bmp_tmp = CreateCompatibleBitmap(hdc_actual, orig_width, orig_height);
SelectObject(hdc, bmp_tmp);
SelectObject(hdc, GetStockObject(BLACK_PEN));
SelectObject(hdc, GetStockObject(BLACK_BRUSH));
Rectangle(hdc, 0, 0, space.right, space.bottom);
// Now adjust usable space to guarantee even division
int cell_width = space.right/board_width;
int cell_height = space.bottom/board_height;
space.right = cell_width * board_width;
space.bottom = cell_height * board_height;
SelectObject(hdc, GetStockObject(WHITE_BRUSH));
Rectangle(hdc, 0, 0, space.right, space.bottom);
// Prepare game piece images
HDC piece1_hdc = CreateCompatibleDC(hdc);
HDC piece2_hdc = CreateCompatibleDC(hdc);
HDC piece3_hdc = CreateCompatibleDC(hdc);
SelectObject(piece1_hdc, piece1);
SelectObject(piece2_hdc, piece2);
SelectObject(piece3_hdc, piece3);
// Setup is complete. Ready to draw.

// Draw grid lines
for (int i=0; i<board_height; i++)
	{
	MoveToEx(hdc, 0, i*cell_height, NULL);
	LineTo(hdc, space.right, i*cell_height);
	}
for (int j=0; j<board_width; j++)
	{
	MoveToEx(hdc, j*cell_width, 0, NULL);
	LineTo(hdc, j*cell_width, space.bottom);
	}

// Fill in pieces
for (int x=0; x<board_width; x++)
	{
	for (int y=0; y<board_height; y++)
		{
		// Determine specifics of this cell
		HDC whichdc;
		uchar player, bombs;
		uchar cell = gameboard->read_cell(x, y);
		player = cell & PLAYER_MASK;
		bombs = cell & BOMB_MASK;
		if (player == PLAYER1)
			whichdc = piece1_hdc;
		else if (player == PLAYER2)
			whichdc = piece2_hdc;

		if ( (replay==3) && (x==last_x) && (y==last_y) )
			{
			whichdc = piece3_hdc;	// last move highlighter
			bombs++;	// force a fake move, to highlight
			}

		// Draw several bombs, depending on how full the cell is
		if (bombs > 0)
			{	// bomb one
			StretchBlt(hdc, x*cell_width+1, y*cell_height+1,
				cell_width/2, cell_height/2, whichdc, 0, 0,
				PIECE_BMP_SIZE, PIECE_BMP_SIZE, SRCCOPY);
			}
		if (bombs > 1)
			{	// bomb two
			StretchBlt(hdc, x*cell_width+(cell_width/2)-1, y*cell_height+1,
				cell_width/2, cell_height/2, whichdc, 0, 0,
				PIECE_BMP_SIZE, PIECE_BMP_SIZE, SRCCOPY);
			}
		if (bombs > 2)
			{	// bomb three
			StretchBlt(hdc, x*cell_width+1, y*cell_height+(cell_height/2)-1,
				cell_width/2, cell_height/2, whichdc, 0, 0,
				PIECE_BMP_SIZE, PIECE_BMP_SIZE, SRCCOPY);
			}
		if (bombs > 3)
			{	// bomb four
			StretchBlt(hdc, x*cell_width+(cell_width/2)-1,
				y*cell_height+(cell_height/2)-1, cell_width/2,
				cell_height/2, whichdc, 0, 0,
				PIECE_BMP_SIZE, PIECE_BMP_SIZE, SRCCOPY);
			}
		}
	}
// We've drawn to memory. Now blast to actual screen.
BitBlt(hdc_actual, 0, 0, orig_width, orig_height, hdc, 0, 0, SRCCOPY);
// Clean up, free GDI resources :)
DeleteDC(piece1_hdc);
DeleteDC(piece2_hdc);
DeleteDC(piece3_hdc);
DeleteObject(bmp_tmp);
DeleteDC(hdc);
ReleaseDC(hWnd, hdc_actual);
}


/*
 Called when we get a time slice, notified by our own timer. This
 may be an opportune time to make that "automatic" move.
*/
void got_timeslice()
{
switch (replay)
	{
	case 0:	// normal operation
		if ( (turn%2 == 0) && (whats_player1 == COMP_PLAYER) )
			computer_move(PLAYER1);
		else if ( (turn%2 == 1) && (whats_player2 == COMP_PLAYER) )
			computer_move(PLAYER2);
		break;

	case 1:
		replay++;	// continue to next stage of replay
		break;

	case 2:
		replay++;	// continue to next stage of replay
		paint_window(NULL);	// shows where the move went
		break;

	case 3:	// last stage of replay
		if (turn%2 == 0)
			gameboard->move_into_cell(PLAYER2, last_x, last_y);
		else
			gameboard->move_into_cell(PLAYER1, last_x, last_y);
		replay = 0;
		paint_window(NULL);
		break;
	}
}

/*
 User has clicked a valid position, already decoded to (x, y)
*/
void valid_click(int x, int y)
{
if (replay != 0)	// if in replay mode
	return;			// don't respond to clicks
if ( (turn%2 == 0) && (whats_player1 == HUMAN_PLAYER) )
	human_move(PLAYER1, x, y);
else if ( (turn%2 == 1) && (whats_player2 == HUMAN_PLAYER) )
	human_move(PLAYER2, x, y);
paint_window(NULL);
winner_check();
}

/*
 User tries to move into position (x, y)
*/
void human_move(uchar player, int x, int y)
{
history.push_back(*gameboard);	// store current state
if (gameboard->move_into_cell(player, x, y))	// attempt move
	turn++;
else	// the move failed, do not store undo state
	history.pop_back();	// remove last stored state
}


/*
 Computer makes either a random, or a smart move
*/
void computer_move(uchar player)
{
history.push_back(*gameboard);	// store current state
if (intelligent)
	{
	SetCursor(hourglass);
	gameboard->smart_move(player);
	SetCursor(pointer);
	}
else
	gameboard->random_move(player);
turn++;
paint_window(NULL);
winner_check();
}


/*
 See if anybody has won.
*/
void winner_check()
{
if (turn > 1)
	{
	uchar victor = gameboard->who_won();
	if (victor == PLAYER1)
		{
		if (sound_on)
			{
			if (whats_player2==COMP_PLAYER)
				PlaySound(FILE_GOODSOUND, NULL, SND_FILENAME );
			else
				PlaySound(FILE_BADSOUND, NULL, SND_FILENAME );
			}
		MessageBox(hWnd, MSG_PLAYER1_WON, TITLE_PROG, MB_ICONASTERISK);
		p1_wins++;
		}
	else if (victor == PLAYER2)
		{
		if (sound_on)
			{
			if (whats_player1==COMP_PLAYER)
				PlaySound(FILE_GOODSOUND, NULL, SND_FILENAME );
			else
				PlaySound(FILE_BADSOUND, NULL, SND_FILENAME );
			}
		MessageBox(hWnd, MSG_PLAYER2_WON, TITLE_PROG, MB_ICONASTERISK);
		p2_wins++;
		}
	}
}


/*
 Delete allocated resources
*/
void cleanup_all()
{
if (gameboard != NULL)
	delete gameboard;
DeleteObject(piece1);
DeleteObject(piece2);
DeleteObject(piece3);
}
